//! miniAMR v1 writes and reads to same memory with an offset to avoid block synchronization.
//!
//! This version can only be used with global memory, but has hopefully a more efficient memory access.

constexpr int XG = 40;
constexpr int YG = 40;
constexpr int ZG = 10;

#pragma imcl grid(input)
#pragma imcl boundary_cond(input:none)
void stencil_kernel(
        int var,
        int num_iter,
        int num_vars,
        Block<double> input
) {
    constexpr int X = XG + 2;
    constexpr int Y = YG + 2;
    constexpr int Z = ZG + 2;

    for (int i = 0; i < num_iter; ++i) {
        const int READ = (i * (num_vars + 1) + (var + 1)) * X * Y * Z;
        const int WRITE = (i * (num_vars + 1) + var) * X * Y * Z;
        // 5d -> 3d
        Block<double> read = input + READ;
        Block<double> write = input + WRITE;
        write[idz + 1][idy + 1][idx + 1] =
                (read[idz][idy + 1][idx + 1]
                 + read[idz + 1][idy][idx + 1]
                 + read[idz + 1][idy + 1][idx]
                 + read[idz + 1][idy + 1][idx + 1]
                 + read[idz + 1][idy + 1][idx + 2]
                 + read[idz + 1][idy + 2][idx + 1]
                 + read[idz + 2][idy + 1][idx + 1]) / 7.0;
    }
}
