//! miniAMR v2 uses separate memory for reading and writing to avoid block synchronization.
//!
//! This version can be used together with global and hopefully constant memory.

constexpr int XG = 40;
constexpr int YG = 40;
constexpr int ZG = 10;

#pragma imcl grid(input)
#pragma imcl boundary_cond(input:none,output:none)
void stencil_kernel(
        int var,
        int num_iter,
        int num_vars,
        Block<double> input,
        Block<double> output
) {
    constexpr int X = XG + 2;
    constexpr int Y = YG + 2;
    constexpr int Z = ZG + 2;

    for (int i = 0; i < num_iter; ++i) {
        const int OFFSET = (i * num_vars + var) * X * Y * Z;
        // 5d -> 3d
        Block<double> read = input + OFFSET;
        Block<double> write = output + OFFSET;
        write[idz + 1][idy + 1][idx + 1] =
                (read[idz][idy + 1][idx + 1]
                 + read[idz + 1][idy][idx + 1]
                 + read[idz + 1][idy + 1][idx]
                 + read[idz + 1][idy + 1][idx + 1]
                 + read[idz + 1][idy + 1][idx + 2]
                 + read[idz + 1][idy + 2][idx + 1]
                 + read[idz + 2][idy + 1][idx + 1]) / 7.0;
    }
}
