// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "type.h"

using namespace std;

Type::Type() {
    this->baseType = INT;
    this->pointerLevel = 0;
}

Type::Type(BaseType baseType) {
    this->baseType = baseType;
    this->pointerLevel = 0;
}

bool Type::isConstantMemCandidate() {
    int size = 1;
    for (int i = 0; i < this->indices.size(); i++) {
        size *= this->indices[i];
        if (size < 0)
            return false;
    }

    return size < 256;
}

BaseType Type::parseType(string typeString) {
    if (typeString == "float") {
        return FLOAT;
    } else if (typeString == "int") {
        return INT;
    } else if (typeString == "uchar") {
        return UCHAR;
    } else if (typeString == "unsigned char") {
        return UCHAR;
    } else if (typeString == "double") {
        return DOUBLE;
    } else {
        return NOTYPE;
    }
}

string Type::baseTypeToString(BaseType baseType) {
    switch (baseType) {
        case FLOAT:
            return "float";
        case INT:
            return "int";
        case UCHAR:
            return "unsigned char";
        case IMAGE2D_T:
            return "image2d_t";
        case IMAGE3D_T:
            return "image3d_t";
        case DOUBLE:
            return "double";
        default:
            return "no base type";
    }
}


string Type::baseTypeToMpiString(BaseType baseType) {
    switch (baseType) {
        case FLOAT:
            return "MPI_FLOAT";
        case INT:
            return "MPI_INT";
        case UCHAR:
            return "MPI_UNSIGNED_CHAR";
        case IMAGE2D_T:
            return "image2d_t";
        case IMAGE3D_T:
            return "image3d_t";
        case DOUBLE:
            return "MPI_DOUBLE";
        default:
            return "no base type";
    }
}
