// ************************************************************************
//
// miniAMR: stencil computations with boundary exchange and AMR.
//
// Copyright (2014) Sandia Corporation. Under the terms of Contract
// DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government 
// retains certain rights in this software.
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
// Questions? Contact Courtenay T. Vaughan (ctvaugh@sandia.gov)
//                    Richard F. Barrett (rfbarre@sandia.gov)
//
// ************************************************************************

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>

#include "proto.h"
#include "timer.h"

double timer(void) {
    return (MPI_Wtime());
}

void *ma_malloc(size_t size, char *file, int line) {
    void *ptr;

    ptr = malloc(size);

    if (ptr == NULL) {
        printf("NULL pointer from malloc call in %s at %d\n", file, line);
        exit(-1);
    }

    counter_malloc++;
    size_malloc += (double) size;

    return (ptr);
}

#if defined(MINIAMR_CUDA)
cudaError_t ma_cudaMalloc(void **devPtr, size_t size) {
    ++gpu_num_alloc_calls;
    gpu_bytes_allocated += size;
    return cudaMalloc(devPtr, size);
}

cudaError_t ma_cudaMemcpy(void *dst, const void *src, size_t count, enum cudaMemcpyKind kind) {
    gpu_bytes_transfered += count;
    return cudaMemcpy(dst, src, count, kind);
}

cudaError_t ma_cudaMalloc3DArray(cudaArray_t *array, const struct cudaChannelFormatDesc* desc, struct cudaExtent extent) {
    ++gpu_num_alloc_calls;
    gpu_bytes_allocated += extent.width * extent.height * extent.depth;
    return cudaMalloc3DArray(array, desc, extent, 0);
}

cudaError_t ma_cudaMemcpy3D(const struct cudaMemcpy3DParms *p) {
    gpu_bytes_transfered += p->extent.width * p->extent.height * p->extent.depth;
    return cudaMemcpy3D(p);
}

cudaError_t ma_cudaMemcpyToSymbol(const void *symbol, const void *src, size_t count, size_t offset, enum cudaMemcpyKind kind) {
    gpu_bytes_transfered += count;
    return cudaMemcpyToSymbol(symbol, src, count, offset, kind);
}
#endif
